@@ -25,32 +25,43 @@ module Yesod.Markdown
   , writePandocTrusted
   -- * Option sets
   , yesodDefaultWriterOptions
-  , yesodDefaultParserState
+  , yesodDefaultReaderOptions
   -- * Form helper
   , markdownField
   )
   where
 
-import Yesod.Form (ToField(..), areq, aopt)
-import Yesod.Core (RenderMessage, SomeMessage(..))
-import Yesod.Form.Types
-import Yesod.Widget (toWidget)
-import Text.Hamlet (hamlet, Html)
+import Data.Monoid (Monoid)
+import Data.String (IsString)
+import Data.Text (Text)
+import Data.Text.Encoding (decodeUtf8With)
+import Data.Text.Encoding.Error (lenientDecode)
+
 import Database.Persist (PersistField)
-
-import Text.Blaze (preEscapedString, preEscapedText)
-import Text.Pandoc
-import Text.HTML.SanitizeXSS (sanitizeBalance)
-
-import Data.Monoid      (Monoid)
-import Data.String      (IsString)
 import System.Directory (doesFileExist)
 
-import qualified Data.Text as T
+import Text.Blaze (ToMarkup (toMarkup))
+import Text.Blaze.Html (preEscapedToMarkup)
+import Text.HTML.SanitizeXSS (sanitizeBalance)
+import Text.Hamlet (hamlet, Html)
+import Text.Pandoc
 
-newtype Markdown = Markdown String
+import Yesod.Core (RenderMessage)
+import Yesod.Form (ToField(..), areq, aopt)
+import Yesod.Form.Functions (parseHelper)
+import Yesod.Form.Types
+import Yesod.Widget (toWidget)
+
+import qualified Data.ByteString as B
+import qualified Data.Text       as T
+
+newtype Markdown = Markdown { unMarkdown :: Text }
     deriving (Eq, Ord, Show, Read, PersistField, IsString, Monoid)
 
+instance ToMarkup Markdown where
+    -- | Sanitized by default
+    toMarkup = markdownToHtml
+
 instance ToField Markdown master where
     toField = areq markdownField
 
@@ -59,77 +70,59 @@ instance ToField (Maybe Markdown) master where
 
 markdownField :: RenderMessage master FormMessage => Field sub master Markdown
 markdownField = Field
-    { fieldParse = blank $ Right . Markdown . unlines . lines' . T.unpack
+    { fieldParse = parseHelper $ Right . Markdown . T.filter (/= '\r')
     , fieldView  = \theId name attrs val _isReq -> toWidget
-        [hamlet|
+        [hamlet|$newline never
 <textarea id="#{theId}" name="#{name}" *{attrs}>#{either id unMarkdown val}
 |]
-     }
+    , fieldEnctype = UrlEncoded
+    }
 
-     where
-        unMarkdown :: Markdown -> T.Text
-        unMarkdown (Markdown s) = T.pack s
-
-        lines' :: String -> [String]
-        lines' = map go . lines
-
-        go []        = []
-        go ('\r':xs) = go xs
-        go (x:xs)    = x : go xs
-
-blank :: (Monad m, RenderMessage master FormMessage)
-      => (T.Text -> Either FormMessage a)
-      -> [T.Text]
-      -> m (Either (SomeMessage master) (Maybe a))
-blank _ []     = return $ Right Nothing
-blank _ ("":_) = return $ Right Nothing
-blank f (x:_)  = return $ either (Left . SomeMessage) (Right . Just) $ f x
-
--- | Converts markdown directly to html using the yesod default option 
---   sets and sanitization.
 markdownToHtml :: Markdown -> Html
 markdownToHtml = writePandoc yesodDefaultWriterOptions
-               . parseMarkdown yesodDefaultParserState
+               . parseMarkdown yesodDefaultReaderOptions
 
--- | Same but with no sanitization run
+-- | No HTML sanitization
 markdownToHtmlTrusted :: Markdown -> Html
 markdownToHtmlTrusted = writePandocTrusted yesodDefaultWriterOptions
-                      . parseMarkdown yesodDefaultParserState
+                      . parseMarkdown yesodDefaultReaderOptions
 
--- | Reads markdown in from the specified file; returns the empty string 
---   if the file does not exist
+-- | Returns the empty string if the file does not exist
 markdownFromFile :: FilePath -> IO Markdown
 markdownFromFile f = do
-    exists <- doesFileExist f
-    content <- do
+    exists  <- doesFileExist f
+    content <-
         if exists
-            then readFile f
+            then readFileUtf8 f
             else return ""
 
     return $ Markdown content
 
--- | Converts the intermediate Pandoc type to Html. Sanitizes HTML.
+    where
+        readFileUtf8 :: FilePath -> IO Text
+        readFileUtf8 fp = do
+            bs <- B.readFile fp
+            return $ decodeUtf8With lenientDecode bs
+
 writePandoc :: WriterOptions -> Pandoc -> Html
-writePandoc wo = preEscapedText . sanitizeBalance . T.pack . writeHtmlString wo
+writePandoc wo = preEscapedToMarkup . sanitizeBalance . T.pack . writeHtmlString wo
 
--- | Skips the sanitization and its required conversion to Text
 writePandocTrusted :: WriterOptions -> Pandoc -> Html
-writePandocTrusted wo = preEscapedString . writeHtmlString wo
+writePandocTrusted wo = preEscapedToMarkup . writeHtmlString wo
 
--- | Parses Markdown into the intermediate Pandoc type
-parseMarkdown :: ParserState -> Markdown -> Pandoc
-parseMarkdown ro (Markdown m) = readMarkdown ro m
+parseMarkdown :: ReaderOptions -> Markdown -> Pandoc
+parseMarkdown ro = readMarkdown ro . T.unpack . unMarkdown
 
--- | Pandoc defaults, plus Html5, minus WrapText
+-- | Defaults plus Html5, minus WrapText
 yesodDefaultWriterOptions :: WriterOptions
-yesodDefaultWriterOptions = defaultWriterOptions
+yesodDefaultWriterOptions = def
   { writerHtml5    = True
   , writerWrapText = False
   }
 
--- | Pandoc defaults, plus Smart, plus ParseRaw
-yesodDefaultParserState :: ParserState
-yesodDefaultParserState = defaultParserState
-    { stateSmart    = True
-    , stateParseRaw = True
+-- | Defaults plus Smart and ParseRaw
+yesodDefaultReaderOptions :: ReaderOptions
+yesodDefaultReaderOptions = def
+    { readerSmart    = True
+    , readerParseRaw = True
     }
